#ifndef __POINT4D_H__
#define __POINT4D_H__

#include "mathlibx.h"

//////////////////////////////////////////////////
// Operations Supported
//
// Point p, q, r;
// Vector u, v, w;   // Vectors
// FLOAT f;          // Scalars
// BOOL b;           // Booleans
//
// f = p[0];       // LHS Index Operation
// p[0] = f;       // LHS Index Operation
// f = p(0);       // RHS Index Operation
// p(0) = f;       // ILLEGAL -- RHS (right hand side) only!!!
//
// p = q;
// p = v;
// p += v;
// p -= v;
// p = q + v;
// v = p - q;
// f = p.Dot(v);   // Dot Product for on Plane calculations

// Operations for Averaging to find Centroids and Doing Linear Combinations
// p += q;
// p *= f;         // Scalar Multiplication
// p /= f;         // Scalar Multiplication
// p = q + r;
// p = q * f;      // Scalar Multiplication
// p = f * q;      // Scalar Multiplication
// p = q / f;      // Scalar Division
// p = f / q;      // Scalar Inversion
//

// Logic
// b = (p == q); // Quasi Equal allows for floating point inaccuracy
// b = (p != q); // Quasi Not Equal allows for floating point inaccuracy

// Bounding Box Checking Operations
// b = (p < q);  // Elementwise
// b = (p <= q); // Elementwise
// b = (p > q);  // Elementwise
// b = (p >= q); // Elementwise

//////////////////////////////////////////////////////////////////////
// Point Class
//

class Vector;

class Point
{
public:
  Point();
  Point(FLOAT x, FLOAT y, FLOAT z, FLOAT w = 1.0f);
  virtual ~Point();

public:
  inline VOID Set(FLOAT x, FLOAT y, FLOAT z, FLOAT w = 1.0f);
  inline VOID Set(const POINT4D pt);
  inline VOID Get(POINT4D pt) const;
  inline const FLOAT *Get() const;

public:
  VOID Null();
  BOOL Homogenize();

public:
  inline FLOAT &operator[](const UINT_32 uiIndex);
  inline FLOAT operator()(const UINT_32 uiIndex) const;

  Point &operator=(const Point &pt);
  Point &operator=(const Vector &v);
  Point &operator+=(const Vector &v);
  Point &operator-=(const Vector &v);

  Point &operator+=(const Point &pt);
  Point &operator*=(const FLOAT f);
  Point &operator/=(const FLOAT f);

  BOOL operator==(const Point &pt) const;
  BOOL operator!=(const Point &pt) const;

  FLOAT Dot(const Vector &v) const;

  BOOL operator<(const Point &pt) const;
  BOOL operator<=(const Point &pt) const;
  BOOL operator>(const Point &pt) const;
  BOOL operator>=(const Point &pt) const;

protected:
  FLOAT Dot(const Point &pt) const; // For Equality Checking

public:
  VOID Display() const;

protected:
  POINT4D m_pt;
};

//////////////////////////////////////////////////////////////////////
// External Operators
//

inline Point operator+(const Point &pt, const Vector &v);
inline Vector operator-(const Point &pt, const Point &q);

inline Point operator+(const Point &pt, const Point &q);
inline Point operator*(const Point &pt, const FLOAT f);
inline Point operator*(const FLOAT f, const Point &pt);
inline Point operator/(const Point &pt, const FLOAT f);

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline VOID Point::Set(FLOAT x, FLOAT y, FLOAT z, FLOAT w/* = 1.0f*/)
{
  m_pt[X] = x;
  m_pt[Y] = y;
  m_pt[Z] = z;
  m_pt[W] = w;
}

inline VOID Point::Set(const POINT4D pt)
{
  ASSERT( pt != NULL );

  memcpy(m_pt, pt, sizeof(POINT4D));
}

inline VOID Point::Get(POINT4D pt) const
{
  ASSERT( pt != NULL );

  memcpy(pt, m_pt, sizeof(POINT4D));
}

inline const FLOAT *Point::Get() const
{
  return m_pt;
}

inline FLOAT &Point::operator[](const UINT_32 uiIndex)
{
  return m_pt[uiIndex];
}

inline FLOAT Point::operator()(const UINT_32 uiIndex) const
{
  return m_pt[uiIndex];
}

//////////////////////////////////////////////////////////////////////
// External Operator Implementations
//

#include "vector.h"

inline Point operator+(const Point &pt, const Vector &v)
{
  return Point(pt(0) + v(0), pt(1) + v(1), pt(2) + v(2), pt(3) + v(3));
}

inline Vector operator-(const Point &pt, const Point &q)
{
  return Vector(pt(0) - q(0), pt(1) - q(1), pt(2) - q(2), pt(3) - q(3));
}

inline Point operator+(const Point &pt, const Point &q)
{
  return Point(pt(0) + q(0), pt(1) + q(1), pt(2) + q(2), pt(3) + q(3));
}

inline Point operator*(const Point &pt, const FLOAT f)
{
  return Point(pt(0)*f, pt(1)*f, pt(2)*f, pt(3)*f);  
}

inline Point operator*(const FLOAT f, const Point &pt)
{
  return Point(pt(0)*f, pt(1)*f, pt(2)*f, pt(3)*f);  
}

inline Point operator/(const Point &pt, const FLOAT f)
{
  ASSERT( f != 0.0 );

  return Point(pt(0)/f, pt(1)/f, pt(2)/f, pt(3)/f);  
}

//////////////////////////////////////////////////////////////////////
// DynPoint Class
//

class DynPoint : public Point
{
public:
  DynPoint();
  DynPoint(FLOAT x, FLOAT y, FLOAT z, FLOAT w = 1.0f);
  virtual ~DynPoint();

public:
  inline VOID SetAll(DynPoint *pdynpt);
  inline VOID SetAll(DynFloat *pdynfX, DynFloat *pdynfY = NULL, 
		     DynFloat *pdynfZ = NULL, DynFloat *pdynfW = NULL);
  inline VOID SetDynExprs(CHAR *pcX, CHAR *pcY = NULL, 
                          CHAR *pcZ = NULL, CHAR *pcW = NULL);
  virtual inline BOOL GetDynamic();

public:
  virtual VOID Update();

protected:
  CHAR *m_ppcDynExpr[4];
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline VOID DynPoint::SetAll(DynPoint *pdynpt)
{
  ASSERT( pdynpt != NULL );

  memcpy(this, pdynpt, sizeof(DynPoint));
}

inline VOID DynPoint::SetAll(DynFloat *pdynfX, DynFloat *pdynfY/* = NULL*/, 
			  DynFloat *pdynfZ/* = NULL*/, 
			  DynFloat *pdynfW/* = NULL*/)
{
  Null();
  if ( pdynfX != NULL )
    {
      m_pt[0] = pdynfX->GetValue();
      m_ppcDynExpr[0] = pdynfX->GetDynExpr();
    }
  if ( pdynfY != NULL )
    {
      m_pt[1] = pdynfY->GetValue();
      m_ppcDynExpr[1] = pdynfY->GetDynExpr();
    }
  if ( pdynfZ != NULL )
    {
      m_pt[2] = pdynfZ->GetValue();
      m_ppcDynExpr[2] = pdynfZ->GetDynExpr();
    }
  if ( pdynfW != NULL )
    {
      m_pt[3] = pdynfW->GetValue();
      m_ppcDynExpr[3] = pdynfW->GetDynExpr();
    }
}

inline VOID DynPoint::SetDynExprs(CHAR *pcX, CHAR *pcY/* = NULL*/, 
				  CHAR *pcZ/* = NULL*/, 
				  CHAR *pcW/* = NULL*/)
{
  m_ppcDynExpr[0] = pcX;
  m_ppcDynExpr[1] = pcY;
  m_ppcDynExpr[2] = pcZ;
  m_ppcDynExpr[3] = pcW;
}

inline BOOL DynPoint::GetDynamic()
{
  return ( m_ppcDynExpr[0] != NULL ||
	   m_ppcDynExpr[1] != NULL ||
	   m_ppcDynExpr[2] != NULL ||
	   m_ppcDynExpr[3] != NULL );
}

#endif // __POINT4D_H__

